#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/ioccom.h>
#include <sys/conf.h>
#include <sys/fcntl.h>
#include <sys/tty.h>
#include <sys/file.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/sysctl.h>
#include <sys/uio.h>

#include <dev/usb/usb.h>
#include <dev/usb/usbhid.h>

#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#include "usbdevs.h"
#include <dev/usb/usb_quirks.h>
#include <dev/usb/hid.h>
#include <dev/usb/uxb360gp_rdesc.h>

/*
 * USB joystick/gamepad implementation for FreeBSD
 * Emmanuel Vadot <elbarto@arcadebsd.org>
 */

#include "ujoy.h"

#ifdef USB_DEBUG
#define DPRINTF(x) if (ujoydebug) printf x
#define DPRINTFN(n,x) if (ujoydebug>(n)) printf x
int ujoydebug = 0;
SYSCTL_NODE(_hw_usb, OID_AUTO, ujoy, CTLFLAG_RW, 0, "USB ujoy");
SYSCTL_INT(_hw_usb_ujoy, OID_AUTO, debug, CTLFLAG_RW,
	   &ujoydebug, 0, "ujoy debug level");
#else
#define DPRINTF(x)
#define DPRINTFN(n,x)
#endif

#define UJOYUNIT(s)	(dev2unit(s)&0x1f)
#define UJOY_BUFSIZE	9

/* Needed for the Xbox360 gamepad */
#define HUG_DPAD_UP	0x90
#define HUG_DPAD_DOWN	0x91
#define HUG_DPAD_RIGHT	0x92
#define HUG_DPAD_LEFT	0x93

enum {
  UJOY_HID = 0,			/* HID usb gamepad/joystick */
  UJOY_X360 = 1			/* Xbox 360 gamepad */
};

struct ujoy_softc {
  device_t sc_dev;		/* base device */
  usbd_interface_handle sc_iface;	/* interface */
  usbd_pipe_handle sc_intrpipe;	/* interrupt pipe */
  int sc_ep_addr;

  u_char *sc_ibuf;
  u_int8_t sc_iid;
  int sc_isize;

  struct hid_location	sc_axis[UJOY_MAXAXIS];
  struct hid_location	sc_hat;
  struct hid_location	*sc_buttons;
  struct hid_location	sc_dpad[4];

  int nbuttons;
  int naxis;
  int nhat;

  int sc_enabled;
  int sc_disconnected;	/* device is gone */
  int sc_type;		/* HID, non-hid etc ... */

  struct ujoy	joybuf;

  char vendor[USB_MAX_STRING_LEN];
  char product[USB_MAX_STRING_LEN];

  int   sc_data;

  int	sc_state;
#define	UJOY_ASLEEP	0x01	/* readFromDevice is waiting */
#define	UJOY_SELECT	0x02	/* select is waiting */
  struct selinfo rsel;

  struct cdev *dev;
};

static d_open_t  ujoy_open;
static d_close_t ujoy_close;
static d_read_t  ujoy_read;
static d_ioctl_t ujoy_ioctl;
static d_poll_t  ujoy_poll;

static struct cdevsw ujoy_cdevsw = {
	.d_version =	D_VERSION,
	.d_flags =	D_NEEDGIANT,
	.d_open =	ujoy_open,
	.d_close =	ujoy_close,
	.d_read =	ujoy_read,
	.d_ioctl =	ujoy_ioctl,
	.d_poll =	ujoy_poll,
	.d_name =	"ujoy",
};

static void ujoy_intr(usbd_xfer_handle xfer,
		      usbd_private_handle priv, usbd_status status);

static device_probe_t ujoy_match;
static device_attach_t ujoy_attach;
static device_detach_t ujoy_detach;

static device_method_t ujoy_methods[] = {
	DEVMETHOD(device_probe,		ujoy_match),
	DEVMETHOD(device_attach,        ujoy_attach),
	DEVMETHOD(device_detach,	ujoy_detach),
	{ 0, 0 }
};

static driver_t ujoy_driver = {
	"ujoy",
	ujoy_methods,
	sizeof(struct ujoy_softc)
};

static devclass_t ujoy_devclass;

static int
ujoy_match(device_t self)
{
  struct ujoy_softc *sc = device_get_softc(self);
  struct usb_attach_arg *uaa = device_get_ivars(self);
  usb_interface_descriptor_t *id;
  int size, ret;
  void *desc;
  usbd_status err;

  if (uaa->iface == NULL)
    return (UMATCH_NONE);

  id = usbd_get_interface_descriptor(uaa->iface);
  if (id == NULL)
    return (UMATCH_NONE);

  /* Non HID joystick */
  if (id->bInterfaceClass != UICLASS_HID)
    {
      if (id->bInterfaceClass == UICLASS_VENDOR &&
	  id->bInterfaceSubClass == UISUBCLASS_XBOX360_CONTROLLER &&
	  id->bInterfaceProtocol == UIPROTO_XBOX360_GAMEPAD)
	{
	  sc->sc_type = UJOY_X360;
	  ret = UMATCH_IFACECLASS;
	}
      else
	ret = UMATCH_NONE;
    }
  else /* HID Joystick */
    {
      err = usbd_read_report_desc(uaa->iface, &desc, &size, M_TEMP);
      if (err)
	return (UMATCH_NONE);

      ret = UMATCH_NONE;
      if (hid_is_collection(desc, size,
			    HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_GAME_PAD)) ||
	  hid_is_collection(desc, size,
			    HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_JOYSTICK)))
	{
	  ret = UMATCH_IFACECLASS;
	  sc->sc_type = UJOY_HID;
	}
      free(desc, M_TEMP);
    }
  return (ret);
}

static int
ujoy_attach(device_t self)
{
  struct ujoy_softc *sc = device_get_softc(self);
  struct usb_attach_arg *uaa = device_get_ivars(self);
  usbd_interface_handle iface = uaa->iface;
  usb_interface_descriptor_t *id;
  usb_endpoint_descriptor_t *ed;
  int size;
  void *desc;
  int i;
  struct hid_location loc_btn;
  u_int32_t flags;
  struct hid_item h;
  struct hid_data *d;
  usbd_status err;

  sc->sc_disconnected = 1;
  sc->sc_iface = iface;

  id = usbd_get_interface_descriptor(iface);
  sc->sc_dev = self;
  ed = usbd_interface2endpoint_descriptor(iface, 0);
  if (!ed) {
    device_printf(self, "could not read endpoint descriptor\n");
    return ENXIO;
  }
  if (UE_GET_DIR(ed->bEndpointAddress) != UE_DIR_IN ||
      UE_GET_XFERTYPE(ed->bmAttributes) != UE_INTERRUPT) {
    device_printf(self, "unexpected endpoint\n");
    return ENXIO;
  }

  if (sc->sc_type == UJOY_X360)
    {
      static uByte reportbuf[] = {1, 3, 2};

      if (device_get_unit(self) < 4)
	reportbuf[2] += device_get_unit(self);

      /* The LEDs on the gamepad are blinking by default, set the led to the # of device. */
      usbd_set_report(uaa->iface, UHID_OUTPUT_REPORT, 0,
		      &reportbuf, sizeof reportbuf);
      
      /* The Xbox 360 gamepad has no report descriptor. */
      size = sizeof uhid_xb360gp_report_descr;
      desc = uhid_xb360gp_report_descr;
    }
  else
    {
      err = usbd_read_report_desc(uaa->iface, &desc, &size, M_TEMP);
      if (err)
	return ENXIO;
    }

  /* Axis */
  sc->naxis = 0;
  for (i = 0; i < UJOY_MAXAXIS; i++)
    {
      if (hid_locate(desc, size, HID_USAGE2(HUP_GENERIC_DESKTOP, (HUG_X + i)),
		     hid_input, &sc->sc_axis[i], &flags))
	{
	  sc->joybuf.axis[i].present = 1;
	  for (d = hid_start_parse(desc, size, 1<<hid_input); hid_get_item(d, &h); )
	    {
	      if (h.kind == hid_input && !(h.flags & HIO_CONST) && h.usage == (HID_USAGE2(HUP_GENERIC_DESKTOP, (HUG_X + i))))
		{
		  sc->joybuf.axis[i].min = h.logical_minimum;
		  sc->joybuf.axis[i].max = h.logical_maximum;
		  break;
		}
	    }
	  hid_end_parse(d);
	  sc->naxis++;
	}
      else
	{
	  sc->joybuf.axis[i].present = 0;
	  sc->joybuf.axis[i].value = 0;
	  sc->joybuf.axis[i].min = 0;
	  sc->joybuf.axis[i].max = 0;
	}
    }

  /* Lookup hat */
  sc->nhat = 0;
  if (hid_locate(desc, size, HID_USAGE2(HUP_GENERIC_DESKTOP, HUG_HAT_SWITCH),
		 hid_input, &sc->sc_hat, &flags))
    sc->nhat = 1;

  /* Lookup buttons */
  for (i = 1; i <= UJOY_MAXBUTTONS; i++)
    if (!hid_locate(desc, size, HID_USAGE2(HUP_BUTTON, i),
 		    hid_input, &loc_btn, 0))
      break;
  sc->nbuttons = i - 1;
  
  sc->sc_buttons = malloc(sizeof(struct hid_location)*sc->nbuttons,
			  M_USBDEV, M_NOWAIT);
  if (!sc->sc_buttons) {
    return ENXIO;
  }

  for (i = 1; i <= sc->nbuttons; i++)
    hid_locate(desc, size, HID_USAGE2(HUP_BUTTON, i),
	       hid_input, &sc->sc_buttons[i-1], 0);

  /* Lookup Xbox360 dpad */
  if (sc->sc_type == UJOY_X360)
    for (i = 0; i < 4; i++)
      hid_locate(desc, size, HID_USAGE2(HUP_GENERIC_DESKTOP, (HUG_DPAD_UP + i)),
		 hid_input, &sc->sc_dpad[i], &flags);

  sc->sc_isize = hid_report_size(desc, size, hid_input, &sc->sc_iid);
  sc->sc_ibuf = malloc(sc->sc_isize, M_USB, M_NOWAIT);
  if (!sc->sc_ibuf) {
    return ENXIO;
  }

  sc->sc_ep_addr = ed->bEndpointAddress;
  sc->sc_disconnected = 0;

  if (sc->sc_type == UJOY_HID)
    free(desc, M_TEMP);

  sc->dev = make_dev(&ujoy_cdevsw, device_get_unit(self),
		     UID_ROOT, GID_OPERATOR,
		     0644, "ujoy%d", device_get_unit(self));
  return 0;
}

static int
ujoy_detach(device_t self)
{
  struct ujoy_softc *sc = device_get_softc(self);

  if (sc->sc_enabled)
    sc->sc_enabled = 0;

  DPRINTF(("%s: disconnected\n", device_get_nameunit(self)));

  free(sc->sc_ibuf, M_USB);

  if (sc->sc_state & UJOY_ASLEEP) {
    sc->sc_state &= ~UJOY_ASLEEP;
    wakeup(sc);
  }
  if (sc->sc_state & UJOY_SELECT) {
    sc->sc_state &= ~UJOY_SELECT;
    selwakeuppri(&sc->rsel, PZERO);
  }

  destroy_dev(sc->dev);

  return 0;
}

void
ujoy_intr(usbd_xfer_handle xfer, usbd_private_handle addr, usbd_status status)
{
  struct ujoy_softc *sc = addr;
  u_char *ibuf;
  int i;
  u_char tmp;

  if (status == USBD_CANCELLED)
    return;

  if (status != USBD_NORMAL_COMPLETION) {
    if (status == USBD_STALLED)
      usbd_clear_endpoint_stall_async(sc->sc_intrpipe);
    if(status != USBD_IOERROR)
      return;
  }

  ibuf = sc->sc_ibuf;

  /* Proceed axis */
  for (i = 0; i < UJOY_MAXAXIS; i++)
    if (sc->joybuf.axis[i].present)
      {
	if (sc->joybuf.axis[i].min == 0 && sc->joybuf.axis[i].max <= 255)
	  sc->joybuf.axis[i].value = (u_char)hid_get_data(ibuf, &sc->sc_axis[i]);
	else if (sc->joybuf.axis[i].min == 0 && sc->joybuf.axis[i].max > 255)
	  sc->joybuf.axis[i].value = (u_int)hid_get_data(ibuf, &sc->sc_axis[i]);
	else
	  sc->joybuf.axis[i].value = hid_get_data(ibuf, &sc->sc_axis[i]);
      }

  /* HAT Switch */
  if (sc->nhat)
    sc->joybuf.hat = hid_get_data(ibuf, &sc->sc_hat);
  else
    sc->joybuf.hat = -1;

  /* read Xbox360 dpad has a hat */
  if (sc->sc_type == UJOY_X360)
    {
      if (hid_get_data(ibuf, &sc->sc_dpad[0]))	/* DPAD Up */
	sc->joybuf.hat = 0;
      if (hid_get_data(ibuf, &sc->sc_dpad[2]))	/* DPAD Right */
	sc->joybuf.hat = 2;
      if (hid_get_data(ibuf, &sc->sc_dpad[1]))	/* DPAD Down */
	sc->joybuf.hat = 4;
      if (hid_get_data(ibuf, &sc->sc_dpad[3]))	/* DPAD Left */
	sc->joybuf.hat = 6;
      if (hid_get_data(ibuf, &sc->sc_dpad[0]) &&
	  hid_get_data(ibuf, &sc->sc_dpad[2]))	/* DPAD Up-Right */
	sc->joybuf.hat = 1;
      if (hid_get_data(ibuf, &sc->sc_dpad[0]) &&
	  hid_get_data(ibuf, &sc->sc_dpad[3]))	/* DPAD Up-Left */
	sc->joybuf.hat = 7;
      if (hid_get_data(ibuf, &sc->sc_dpad[1]) &&
	  hid_get_data(ibuf, &sc->sc_dpad[2]))	/* DPAD Down-Right */
	sc->joybuf.hat = 3;
      if (hid_get_data(ibuf, &sc->sc_dpad[1]) &&
	  hid_get_data(ibuf, &sc->sc_dpad[3]))	/* DPAD Down-Left */
	sc->joybuf.hat = 5;
    }

  /* Proceed buttons */
  sc->joybuf.buttons = 0;
  for (i = 0; i < sc->nbuttons; i++)
    {
      tmp = hid_get_data(ibuf, &sc->sc_buttons[i]);
      sc->joybuf.buttons |= ((tmp & 0x1) << i);
    }

  sc->sc_data = 1;

  if (sc->sc_state & UJOY_ASLEEP) {
    sc->sc_state &= ~UJOY_ASLEEP;
    wakeup(sc);
  }
  if (sc->sc_state & UJOY_SELECT) {
    sc->sc_state &= ~UJOY_SELECT;
    selwakeuppri(&sc->rsel, PZERO);
  }
}

static int
ujoy_open(struct cdev *dev, int flag, int fmt, struct thread *p)
{
  struct ujoy_softc *sc;
  usbd_status err;

  sc = devclass_get_softc(ujoy_devclass, UJOYUNIT(dev));
  if (sc == NULL)
    return (ENXIO);

  if (sc->sc_enabled)
    return EBUSY;

  sc->sc_enabled = 1;
  sc->sc_data = 0;

  /* Set up interrupt pipe. */
  err = usbd_open_pipe_intr(sc->sc_iface, sc->sc_ep_addr,
			    USBD_SHORT_XFER_OK, &sc->sc_intrpipe, sc,
			    sc->sc_ibuf, sc->sc_isize, ujoy_intr,
			    USBD_DEFAULT_INTERVAL);
  if (err) {
    DPRINTF(("ujoy_enable: usbd_open_pipe_intr failed, error=%d\n",
	     err));
    sc->sc_enabled = 0;
    return (EIO);
  }

  return 0;
}

static int
ujoy_close(struct cdev *dev, int flag, int fmt, struct thread *p)
{
  struct ujoy_softc *sc;

  sc = devclass_get_softc(ujoy_devclass, UJOYUNIT(dev));
  if (!sc)
    return 0;

  usbd_abort_pipe(sc->sc_intrpipe);
  usbd_close_pipe(sc->sc_intrpipe);
  sc->sc_intrpipe = 0;

  if (sc->sc_enabled)
    sc->sc_enabled = 0;
  return 0;
}

static int
ujoy_read(struct cdev *dev, struct uio *uio, int flag)
{
  struct ujoy_softc *sc;
  int s;
  int error;

  sc = devclass_get_softc(ujoy_devclass, UJOYUNIT(dev));

  s = splusb();
  if (!sc) {
    splx(s);
    return EIO;
  }

  while (sc->sc_data == 0 )
    {
      if (flag & O_NONBLOCK) {		/* non-blocking I/O */
	splx(s);
	return EWOULDBLOCK;
      }

      sc->sc_state |= UJOY_ASLEEP;	/* blocking I/O */
      error = tsleep(sc, PZERO | PCATCH, "ujoyrea", 0);
      if (error) {
	splx(s);
	return error;
      } else if (!sc->sc_enabled) {
	splx(s);
	return EINTR;
      }
      
      sc = devclass_get_softc(ujoy_devclass, UJOYUNIT(dev));
      if (!sc) {
	splx(s);
	return EIO;
      }
    }

  splx(s);
  uiomove(&sc->joybuf, sizeof(struct ujoy), uio);
  s = splusb();
  splx(s);

  sc->sc_data = 0;

  return 0;
}

static int
ujoy_poll(struct cdev *dev, int events, struct thread *p)
{
  struct ujoy_softc *sc;
  int revents = 0;
  int s;

  sc = devclass_get_softc(ujoy_devclass, UJOYUNIT(dev));
  if (!sc)
    return 0;

  s = splusb();
  if (events & (POLLIN | POLLRDNORM)) {
    if (sc->sc_data) {
      revents = events & (POLLIN | POLLRDNORM);
    } else {
      sc->sc_state |= UJOY_SELECT;
      selrecord(p, &sc->rsel);
    }
  }
  splx(s);

  return revents;
}

int
ujoy_ioctl(struct cdev *dev, u_long cmd, caddr_t addr, int flag, struct thread *p)
{
  struct ujoy_softc *sc;

  sc = devclass_get_softc(ujoy_devclass, UJOYUNIT(dev));
  if (!sc)
    return EIO;

  switch(cmd) {
  case UJOY_GETNUMAXIS:
    *(int *)addr = sc->naxis;
    break;
  case UJOY_GETNUMBUTTONS:
    *(int *)addr = sc->nbuttons;
    break;
  }
  return 0;
}

MODULE_DEPEND(ujoy, usb, 1, 1, 1);
DRIVER_MODULE(ujoy, uhub, ujoy_driver, ujoy_devclass, usbd_driver_load, 0);
